/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#include "minrtsFrameData.h"
#define  DEBUG
#define  SMOOTHX  5 
#define maxshift  4
#define DSKYUT    6
#define DSKYAT    4

float minrtsQPk(int nx, float *x, float *y, float *ymax, int neg);
/**********************************************************
 * Usage:
 *    oirCrossCoeff prefix -mask maskFile -cross outFile [-autoshift] [-sky] [-dSky dSky]
 *
 *    prefix:  the basename of oirSci2HiPhot for AOPEN/BOPEN photometry 
 *    on a bright star or sky
 *
 *    maskFile : describes which area to use for estimation
 *
 *    outFile: Output is a IMAGING_DATA table in a fits file with 4 rows
 *    each with only one DATA array 
 *
 *    1:  I1/A
 *    2:  I1/B
 *    3:  I2/A
 *    4:  I2/B
 *
 *    [-autoshift] shift mask to peak of star emission (not valid for -sky)
 *    [-sky]   use sky emission rather than star to estimate crosstalk
 *    [-dSky] try to do an even better sky subtraction from star using
 *         regions spaced dSky pixels farther than default away
 *         from the nominal positions given by the mask.
 ************************************************************* */
void Usage() {
   printf("Usage:\n");
   printf("   oirCrossCoeff prefix -mask maskFile -cross outFile [-autoshift] [-sky] [-dSky dSky]\n\n");
   printf("   prefix  = leading characters of output of oirSci2HiPhot \n");
   printf("             for AOPEN/BOPEN photometry files \n");
   printf("   mask    = MIDI data/detector file containing selection mask\n");
   printf("   outfile = output file name\n");
   printf("   -autoshift (optional): shift mask to peak emission for stars\n");
   printf("   -sky (optional): use sky emission instead of star to estimate coefficients\n");
   printf("   -dSky (opt.); make a new attempt to remove sky background from star\n");
   printf("      value is displancement of sky region away from default\n");
   printf("  \n");
}

/*
   float pixel copying utility
*/
void pixCpy (float *dest, float* source, long nPix) {
   memcpy (dest, source, nPix*sizeof(float));
   return;
}  /* end pixCpy */

/************************************************************
 *     determine best shift of mask by finding maximum signal
 *     empirically
 *******************************************************************/
   void findMaskShift(float *mask1, float *mask2, 
      float *A1, float *A2, float *B1, float *B2, int nx, int ny, float *ABshift) {
      float Acorr[2*maxshift+1], Bcorr[2*maxshift+1];
      float shift[2*maxshift+1];
      double dA, dB;
      float x;
      int i,ix, iy, iShift, is;
      long xy, sy;
      for (iShift=(-maxshift);iShift<=maxshift;iShift++) {
         is = iShift + maxshift;
         shift[is] = iShift;
         Acorr[is] = 0.;
         Bcorr[is] = 0.;
         for (iy=maxshift;iy<(ny-maxshift);iy++) {
            xy = iy*nx;
            sy = (iy + iShift)*nx;
            for (ix=0;ix<nx;ix++) {
               Acorr[is] = Acorr[is] 
                  + A1[ix+xy]*mask1[ix+sy] 
                  + A2[ix+xy]*mask2[ix+sy] ;
               Bcorr[is] = Bcorr[is] 
                  + B1[ix+xy]*mask1[ix+sy] 
                  + B2[ix+xy]*mask2[ix+sy];
            }
         } /* yloop */
      } /* iShift loop */
      ABshift[0] = minrtsQPk(2*maxshift+1, shift, Acorr, &x, 0);
      ABshift[1] = minrtsQPk(2*maxshift+1, shift, Bcorr, &x, 0);
      printf(" For A and B data, shift peaks are %f %f\n",ABshift[0], ABshift[1]);
      return ;
   }  /* end findMaskShift
/**********************************************************
   shift each of the nRegion masks up (if shift >0) or down the given number of
   pixels.  
***********************************************************/

void shiftFrames(minrtsFrameData *frame, float shift) {

   int nRegion, iRegion, nx, ny, ix, iy, oy;
   float *oldy, *newy;
   float *olddata, *newdata;

   nRegion = frame->nregion;
   nx      = frame->nx[0];
   ny      = frame->ny[0];

   oldy    = (float *)malloc(ny * sizeof(float));
   newy    = (float *)malloc(ny * sizeof(float));
   olddata = (float *)malloc(ny * sizeof(float));

   if (shift > (ny-3)) shift = ny-3;
   if (shift < (3-ny)) shift = 3-ny;
   for (iy=0;iy<ny;iy++) {
      oldy[iy] = (float)iy;
      newy[iy] = oldy[iy] - shift;
   }
   if (shift == 0.) return;
   for (iRegion=0;iRegion<nRegion; iRegion++) {
      for (ix=0;ix<nx;ix++) {
         for(iy=0;iy<ny;iy++) olddata[iy] = frame->data[iRegion][ix+nx*iy];
         newdata = minrtsSpline(ny, oldy, olddata, ny, newy, .1);
         for(iy=0;iy<ny;iy++) frame->data[iRegion][ix+nx*iy] = newdata[iy];
         free(newdata);
      }
   }  /* iRegion */
   free(oldy);
   free(newy);
   free(olddata);
   return;
}   /* end of shiftFrame */

int main (int argc, char** argv) {

   char opt1[12];
   char gris[12];
   char shutter[12];

   int iFile;
   int i,ierr, iregion;
   int nx, ny;
   long  nPix;
   int ix, iy;
   int iKey;
   float y;

   int chan1, chan2;

   char outFileName[130];
   char  inFileName[130];
   char *maskFileName;
   char *mFileName;
   char *bang="!";

   char *fileNames[4] = {"A.Aphotometry.fits",
                         "B.Bphotometry.fits",
                         ".Achop.fits",
                         ".Bchop.fits"};

   fitsfile *inFile;
   fitsfile *outFile = NULL;


   imaging_fdata_table_def *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *outData;
   int nRegion;

   char history[80];

   minrtsFrameData *inputFrames[4];
   minrtsFrameData *compressedFrames[4];
   minrtsFrameData *mask, *mask0;

   int inRow;
   float *smWork;
   int iarg;
   float abShift[2] ;
   int useSky    = 0;
   int autoShift = 0;
   int newSky    = 0;
   float dSky    = 0.;
   ierr          = 0;
   maskFileName  = NULL;
   mFileName  = NULL;
   outFileName[0] =0;
/*
      see if nDrop is specified externally
*/
   if (argc < 4) {
      Usage();
      return 1;
   }

/*
       check for sky option
*/
   iarg = 1;
   while (iarg < argc) {
      if (!strcmp("-cross", argv[iarg])) {
         strcpy(outFileName, bang);
         strcat(outFileName, argv[++iarg]);
      }
      if (!strcmp("-mask", argv[iarg])) maskFileName = argv[++iarg];
      if (!strcmp("-m0", argv[iarg]))   mFileName = argv[++iarg];
      if (!strcmp("-autoshift", argv[iarg])) autoShift = 1;
      if (!strcmp("-sky", argv[iarg]))       useSky = 1;
      if (!strcmp("-dSky", argv[iarg])) {
         newSky = 1;
         sscanf(argv[++iarg], " %f ", &dSky);
      }
      ++iarg;
   }
   printf("      ****** oirCrossCoeff *****\n\n"); 
   if (autoShift == 1)  {
      if(useSky == 1) {
         printf("   -autoshift AND -sky CANNOT BE USED SIMULTANEOUSLY\n");
         return 1;
      }
      if(maskFileName == NULL) {
         printf("   -autoshift option requires MASK specification\n");
         return 1;
      }
   }
   if (outFileName[0] == 0) {
      printf("   You need to specify -cross outFileName\n");
      return 1;
   }
   if (useSky) inRow = 7; else inRow = 1;
   if (autoShift) printf("Determining maskShift by crosscorrelation.\n");

/*
       get detector table with coordinate distortion info
*/
/*        
       Loop over input files, print some info
*/
   for (iFile = 0;iFile<4; iFile++) {

/*
      open file and get a lot of parameters for iFile=1
*/
      strcpy(inFileName, argv[1]);
      strcat(inFileName, fileNames[iFile]);
      if (iFile == 0) {
         i = fits_open_file(&inFile, inFileName, READONLY, &ierr) ;
         if (ierr != 0) {
            printf("cfitsio error %i opening file %s\n", ierr, inFileName);
            printf("%s \n",strerror(errno));
            return 1;
         }
         i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", 
            opt1,0, &ierr);
         if (strncmp(opt1, "SCI_PHOT",8) ) {
            printf("I dont handle OPT1 = %s\n", opt1);
            return 1;
         }
         i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", 
            gris,0,&ierr);
         printf("disperser is = %s\n", gris);
/*
      prepare output file
*/

         fits_create_file (&outFile, outFileName, &ierr);
         if (ierr !=0) {
            fprintf (stderr, "error opening fits output %s %i\n", outFileName, ierr);
            printf("%s \n",strerror(errno));
            return 1;
         }
         i = fits_copy_header(inFile, outFile, &ierr);
         snprintf(history,80,"oirCrossCoeff input data %s", argv[1]);
         i = fits_write_history(outFile, history, &ierr);
/*
      move to detector section
*/
         DetectorDef = read_imaging_detector_table(inFile, &ierr);
         if (ierr !=0) {
            printf("error %i opening detector table in %s\n", 
               ierr, argv[1]);
            return 1;
         }
         nRegion = DetectorDef->nregion;
         for (i=0;i<4;i++) DetectorDef->naxis[i][1] = 1; 
         odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
         if (ierr !=0) {
            fprintf (stderr, "error creating data definition\n");
            return 1;
         }
         outData = make_imaging_fdata_row(odata);
/*
     prepare output data structures
*/

         i = fits_close_file(inFile, &ierr);
      }  /* iFile = 0 */
/*
          now process the real data
*/
      
      strcpy(inFileName, argv[1]);
      strcat(inFileName, fileNames[iFile]);
      printf("   Extracting data from File %s\n", inFileName);
      inputFrames[iFile] = minrtsFrameDataReadFitsRow(inFileName, inRow, &ierr);
      if (ierr != 0) {
         printf("Failed to process input file %s: error: %i\n", inFileName, ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
   }   /* loop over inputFiles */
   nx = inputFrames[0]->nx[0];
   ny = inputFrames[0]->ny[0];
   nPix = nx*(long)ny;
/*
      if useSky specified, estimate dark from middle rows of closed apertures
      these are region 1 for A- and region 4 for B-
*/
#ifdef DEBUG
   minrtsFrameDataWriteFitsArray(inputFrames, 4, "input0.fits", &ierr);
#endif
   if (useSky == 1) {
      printf("   Using SKY flux to estimate cross coupling coefficients\n");
      for (ix=0;ix<nx;ix++) {
         y = 0.;
         for (iy=10;iy<30;iy++) y = y + inputFrames[2]->data[0][ix+nx*iy];
         y = y/20.;
         for (iFile=0;iFile<3;iFile=iFile+2)
         for (iregion=1;iregion<3;iregion++) for (iy=0;iy<ny;iy++) 
            inputFrames[iFile]->data[iregion][ix+nx*iy]=
               inputFrames[iFile]->data[iregion][ix+nx*iy] - y;
         y = 0.;
         for (iy=10;iy<30;iy++) y = y + inputFrames[1]->data[3][ix+nx*iy];
         y = y/20.;
         for (iFile=1;iFile<4;iFile=iFile+2)
         for (iregion=1;iregion<3;iregion++) for (iy=0;iy<ny;iy++) 
            inputFrames[iFile]->data[iregion][ix+nx*iy]=
               inputFrames[iFile]->data[iregion][ix+nx*iy] - y;
      }
   } else printf("   Using STAR flux to estimate cross coupling coefficients\n");
/*
      read mask for compression
*/
   if (maskFileName != NULL) {
      mask = minrtsFrameDataReadFits(maskFileName, &ierr);
      if (ierr != 0) {
         printf("ERROR reading mask files %s %i", maskFileName, ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
   } else {
      mask = minrtsFrameDataClone(inputFrames[0]);
      for (iregion=1;iregion<3;iregion++) 
         for (ix=0;ix<nPix;ix++) mask->data[iregion][ix] = 1.;
   }
   if (mFileName != NULL) {
      mask0 = minrtsFrameDataReadFits(mFileName, &ierr);
      if (ierr != 0) {
         printf("ERROR reading m0 files %s %i", mFileName, ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
   } else mask0 = mask;
   if (autoShift != 0) {
/*            call find shift routine
 *            for A- (file 2 and 0) and B- (file 1 and 3)
 *            call minrtsFrameDataCompress with shifted versions of the
 *            mask
 */
      findMaskShift(mask0->data[1], mask0->data[2],
         inputFrames[2]->data[1], inputFrames[2]->data[2], 
         inputFrames[3]->data[1], inputFrames[3]->data[2], nx, ny, abShift) ;
         shiftFrames(inputFrames[0], abShift[0]);
         shiftFrames(inputFrames[2], abShift[0]);
         shiftFrames(inputFrames[1], abShift[1]);
         shiftFrames(inputFrames[3], abShift[1]);
   } 
   for (iFile=0;iFile<4;iFile++) 
      compressedFrames[iFile] = minrtsFrameDataCompress (inputFrames[iFile], mask, 1);
#ifdef DEBUG
   minrtsFrameDataWriteFitsArray(inputFrames, 4, "input.fits", &ierr);
   minrtsFrameDataWriteFitsArray(compressedFrames, 4, "compressed.fits", &ierr);
#endif
#ifdef SMOOTHX
/*
      box smooth channels
*/
   smWork = (float *)malloc(nx * sizeof(float));
   if (SMOOTHX != 0) {
      for (iFile=0;iFile<4;iFile++) 
      for (iregion = 2; iregion<3; iregion++) {
         minrtsBoxSmooth(nx, compressedFrames[iFile]->data[iregion], smWork, SMOOTHX);
         pixCpy (compressedFrames[iFile]->data[iregion], smWork, nx);
         minrtsBoxSmooth(nx, compressedFrames[iFile]->data[iregion], smWork, SMOOTHX);
         pixCpy (compressedFrames[iFile]->data[iregion], smWork, nx);
      }
   }
#endif
#ifdef DEBUG
   minrtsFrameDataWriteFitsArray(compressedFrames, 4, "compresseds.fits", &ierr);
#endif
/*
      truncate photometry channels on bottom at 1% of peak
*/
  for (iFile=0;iFile<2;iFile++) for (iregion=2;iregion<3;iregion++) {
     y = 0.;
     for (ix=0;ix<compressedFrames[iFile]->nx[iregion];ix++) 
     if (compressedFrames[iFile]->data[iregion][ix] > y) 
        y=compressedFrames[iFile]->data[iregion][ix];
     y = 0.01* y;
     if (y > 0) for (ix=0;ix<compressedFrames[iFile]->nx[iregion];ix++) 
     if (compressedFrames[iFile]->data[iregion][ix] < y) 
        compressedFrames[iFile]->data[iregion][ix]=y;
   }
   for (ix=0;ix<compressedFrames[0]->nx[0];ix++) {
      outData->data[0][ix] = compressedFrames[2]->data[1][ix]/compressedFrames[0]->data[1][ix];
      outData->data[1][ix] = compressedFrames[3]->data[1][ix]/compressedFrames[1]->data[1][ix];
      outData->data[2][ix] = compressedFrames[2]->data[2][ix]/compressedFrames[0]->data[2][ix];
      outData->data[3][ix] = compressedFrames[3]->data[2][ix]/compressedFrames[1]->data[2][ix];
   }
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);
/*
      clean up
*/
/*   i = fits_close_file(outFile, &ierr); */
   i = close_imaging_fdata_table(outFile, odata, &ierr); 
   kill_imaging_fdata_row(outData); 
   for (iFile=0;iFile<4;iFile++) {
      minrtsFrameDataKill (inputFrames[iFile]);
      minrtsFrameDataKill (compressedFrames[iFile]);
   }
   if (mask0 != mask) minrtsFrameDataKill(mask0);
   minrtsFrameDataKill (mask);
#ifdef SMOOTHX
   free(smWork);
#endif

   return 0;
}
